/*
 * Copyright (C) 2008 Alan Nisota
 *
 * This file is part of xine, a free video player.
 *
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA
 *
 * Support for dshowserver, which enables using w32 codecs in a seperate process.
 * This is useful to run 32bit codecs in a 64bit xine.
 * It is mostly tageted at supporting CoreAVC at the moment.
 */
/*
 * Modified by Petri Hintukainen <phintuka@users.sourceforge.net> (for xineliboutput) 2008-08-04:
 *
 *  - Modified to build outside of xine-lib source tree (+ Makefile added)
 *  - Added support for xine-lib 1.2 branch
 *  - Modified PTS handling
 *  - Some buffer flag handling fixes
 *  - Added BUF_FLAG_ASPECT handling
 *  - Use xine_fast_memcpy
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>

#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>

#include <semaphore.h>
#include <pthread.h>
#include <sys/mman.h>

#include <xine/xine_internal.h>
#include <xine/video_out.h>
#include <xine/buffer.h>
#include <xine/xineutils.h>

#include "nal_parser.h"

struct vd_struct {
  union {
    uint32_t ret;
    uint32_t cmd;
  };
  uint32_t buflen;
  uint64_t pts;
  uint32_t unused[8];
} __attribute__((__packed__));

enum {
  VD_END = 1,
  VD_DECODE = 2,
  VD_SEEK = 3,
  VD_RELOAD_BIH = 4,
  VD_HAS_BIH = 0x10000,
  VD_VERSION_MASK = 0xFFFF,
};

typedef struct {
    int fd;
    void *mem;
    char *data;
    char *picture;
    char *pagestart;
    int picsize;
    int pagesize;
    sem_t *sem_rd;
    sem_t *sem_wr;
    struct vd_struct *vd;
} ds_mpi_t;

typedef struct {
  video_decoder_class_t   decoder_class;
} dshowserver_class_t;

typedef struct {
  video_decoder_t   video_decoder;

  xine_stream_t    *stream;
  dshowserver_class_t *class;
  
  int64_t           video_step;
  int64_t           pts;
  int               decoder_ok;
  int               decoder_initialized;

  int               stream_id;
  int               skipframes;

  int               size;        /* the current size of buf       */
  int               width;       /* the width of a video frame    */
  int               height;      /* the height of a video frame   */
  uint32_t          compression; /* the compression format */
  double            ratio;       /* the width to height ratio     */
  int               bpp;
  int               numpages;
  int               extra;       /* width & 0x0f */
  ds_mpi_t          ds_mpi;

  xine_bmiheader    *bih;
  unsigned char     *extradata;
  int               extradata_size;
  struct nal_parser *parser;     /* Parser for H264 */

} dshowserver_decoder_t;

static int sem_twait(sem_t *sem, int t) {
    struct timespec ts;
    clock_gettime(CLOCK_REALTIME, &ts);
    ts.tv_sec += t;
    return(sem_timedwait(sem, &ts));
}

static void make_bih(dshowserver_decoder_t *this)
{
  if(this->bih) {
    this->bih = realloc(this->bih, sizeof(xine_bmiheader) + this->extradata_size);
  } else {
    this->bih = calloc(1, sizeof(xine_bmiheader) + this->extradata_size);
    this->bih->biWidth = this->width;
    this->bih->biHeight = this->height;
    this->bih->biPlanes = 1;
    this->bih->biBitCount = 24;
    this->bih->biCompression = this->compression;
    this->bih->biSizeImage = 0;
    this->bih->biXPelsPerMeter=10000;
    this->bih->biYPelsPerMeter=10000;
    this->bih->biClrUsed=0;
    this->bih->biClrImportant=0;
  }
  if(this->extradata_size)
    memcpy((unsigned char *)this->bih + sizeof(xine_bmiheader), this->extradata, this->extradata_size);
  this->bih->biSize = sizeof(xine_bmiheader) + this->extradata_size;
}

static void dshowserver_decode_data (video_decoder_t *this_gen, buf_element_t *buf) {
  dshowserver_decoder_t *this = (dshowserver_decoder_t *) this_gen;
  int ret                               = 0;
  vo_frame_t *img                       = 0; /* video out frame */
  int size_from_nal                     = 0;

  lprintf ("processing packet type = %08x, buf->decoder_flags=%08x\n", 
	   buf->type, buf->decoder_flags);

  if (buf->decoder_flags & BUF_FLAG_FRAMERATE) {
    this->video_step = buf->decoder_info[0];
    _x_stream_info_set(this->stream, XINE_STREAM_INFO_FRAME_DURATION, this->video_step);
  
    lprintf ("video_step is %lld\n", this->video_step);
  }

  if (buf->decoder_flags & BUF_FLAG_PREVIEW)
    return;

  if (buf->decoder_flags & BUF_FLAG_SPECIAL) {
    /* store extradata */
    if (buf->decoder_info[1] == BUF_SPECIAL_DECODER_CONFIG &&
        !this->extradata_size)
    {
      lprintf("BUF_SPECIAL_DECODER_CONFIG\n");
      this->extradata_size = buf->decoder_info[2];
      this->extradata = calloc(1, buf->decoder_info[2]);
      memcpy(this->extradata, buf->decoder_info_ptr[2],
              buf->decoder_info[2]);
      lprintf("Extradata: %d\n", this->extradata_size);
    }
  }
    
  if(! this->decoder_ok && ! (buf->decoder_flags & BUF_FLAG_STDHEADER)) {
    int len = 0;
    unsigned char *tmpbuf;
    int tmplen;
    if(! this->parser)
      this->parser = init_parser();

    while(len < buf->size) {
      len += parse_frame(this->parser, buf->content + len, buf->size - len,
                         &tmpbuf, &tmplen);
      if(tmpbuf)
        free(tmpbuf);
      if(this->parser->current_nal->sps != NULL)
      {
	this->width = this->parser->current_nal->sps->pic_width;
	this->height = this->parser->current_nal->sps->pic_height;
	this->compression = 0x34363248; //H264
        size_from_nal = 1;
        printf("size_from_nal: %d x %d\n", this->width, this->height);
      }
    }
  }

/* TODO: split to  if(BUF_FLAG_HEADER) proc_header_data(); else proc_data(); */

  if (! this->decoder_ok && (buf->decoder_flags & BUF_FLAG_STDHEADER || size_from_nal)) {
    if(! size_from_nal) {
      int size = ((xine_bmiheader *) buf->content)->biSize;
      this->bih = calloc(1, size);
      memcpy(this->bih, buf->content, size);
      this->width                         = this->bih->biWidth;
      this->height                        = this->bih->biHeight;
      this->compression                   = this->bih->biCompression;
    }
    this->extra                         = this->width & 0x0F;
    this->bpp                           = 16; //YUY2
    this->numpages                      = 10;

    if ( buf->type & 0xff )
      return;
    
    lprintf ("processing header ...\n"); 

    /* init package containing bih */
    if (buf->decoder_flags & BUF_FLAG_ASPECT)
      this->ratio = (double)buf->decoder_info[1] / (double)buf->decoder_info[2];
    else
      this->ratio = (double)this->width/(double)this->height;

    _x_stream_info_set(this->stream, XINE_STREAM_INFO_VIDEO_WIDTH,    this->width);
    _x_stream_info_set(this->stream, XINE_STREAM_INFO_VIDEO_HEIGHT,   this->height);

    //Do dshowserver init here
    this->decoder_ok = 1;

    if (! this->decoder_ok) {
      xine_log (this->stream->xine, XINE_LOG_MSG,
		_("dshowserver: decoder failed to start\n"));
      _x_stream_info_set(this->stream, XINE_STREAM_INFO_VIDEO_HANDLED, 0);
      _x_message(this->stream, XINE_MSG_LIBRARY_LOAD_ERROR,
                   "dshowserver", NULL);
    }
  } else if (this->decoder_ok) {
    lprintf ("processing packet ...\n"); 
    if(! this->decoder_initialized) {
/* TODO: make this separate static function */
      //delay initialization until 1st frame in case extradata comes.
      int memsize                         = 0;
      char cmd[255], shm[80], sem1[80], sem2[80];

      this->decoder_initialized = 1;
      snprintf(cmd, 255, "dshowserver -c CoreAVCDecoder.ax -s %dx%d "
               "-g 09571a4b-f1fe-4c60-9760de6d310c7c31 "
               "-f 0x%08x -b %d -o 0x%08x -p %d -i %x -n %d %s&",
               this->width, this->height,
               this->compression, this->bpp, XINE_IMGFMT_YUY2, getpid(), *(int *)pthread_self(), this->numpages, "");
      snprintf(shm, 80, "/dshow_shm.%x", *(int *)pthread_self());
      snprintf(sem1, 80, "/dshow_sem1.%x", *(int *)pthread_self());
      snprintf(sem2, 80, "/dshow_sem2.%x", *(int *)pthread_self());
      this->ds_mpi.fd = shm_open(shm, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
      this->ds_mpi.picsize =  this->width * this->height * this->bpp / 8;
      this->ds_mpi.pagesize = this->ds_mpi.picsize + 1024;
      memsize = sizeof(struct vd_struct) + this->width * this->height + this->ds_mpi.picsize + this->extra + this->ds_mpi.pagesize * this->numpages;
      ftruncate(this->ds_mpi.fd, memsize);
      this->ds_mpi.mem = mmap(NULL, memsize, PROT_READ | PROT_WRITE, MAP_SHARED, this->ds_mpi.fd, 0);
      if(this->ds_mpi.mem == MAP_FAILED) {
        perror("mmap");
        return;
      }
      memset((char *)this->ds_mpi.mem, 0, sizeof(struct vd_struct));
      this->ds_mpi.vd = (struct vd_struct *)this->ds_mpi.mem;
      this->ds_mpi.data = ((char *)this->ds_mpi.mem) + sizeof(struct vd_struct);
      this->ds_mpi.picture = this->ds_mpi.data + this->width * this->height;
      this->ds_mpi.pagestart = this->ds_mpi.picture + this->ds_mpi.picsize + this->extra;
      if (this->extra)
        memset((char *)this->ds_mpi.picture + this->ds_mpi.picsize, 0, this->extra);
  
      if(this->extradata)
        make_bih(this);
      if(this->bih && this->bih->biSize > sizeof(xine_bmiheader)) {
        this->ds_mpi.vd->cmd |= VD_HAS_BIH; //Use embedded bih
        memcpy(this->ds_mpi.data, this->bih, this->bih->biSize);
      }
      //Create read/write semaphores in locked state
      this->ds_mpi.sem_wr = sem_open(sem1, O_CREAT, 0644, 0);
      this->ds_mpi.sem_rd = sem_open(sem2, O_CREAT, 0644, 0);
      system(cmd);
      ret = sem_twait(this->ds_mpi.sem_rd, 10);
      shm_unlink(shm);
      sem_unlink(sem1);
      sem_unlink(sem2);
      if(ret != 0) {
        printf("DirectShow filter failed");
        return;
      }
      printf("Found DirectShow filter");
    }

    if( this->stream_id < 0 )
       this->stream_id = buf->type & 0xff;

    if( this->stream_id != (buf->type & 0xff) )
       return;

    /* do not try to decode header buffers */
    if (buf->decoder_flags & BUF_FLAG_HEADER)
      return;
  
    if (buf->decoder_flags & BUF_FLAG_FRAME_START)
      this->size = 0;
  
    if (buf->size > 0) {
      xine_fast_memcpy (this->ds_mpi.data + this->size, buf->content, buf->size);
      this->size += buf->size;
    }

    /* store pts from first packet, it is not carried in last chunk */
    if (buf->pts > 0)
      this->pts = buf->pts;
  
    if (! (buf->decoder_flags & BUF_FLAG_FRAME_END))
      return;

    this->ds_mpi.vd->cmd = VD_DECODE; //'2' is cmd for decoding
    this->ds_mpi.vd->pts = this->pts;
    this->ds_mpi.vd->buflen = this->size;
    this->size = 0;
    sem_post(this->ds_mpi.sem_wr);
    ret = sem_twait(this->ds_mpi.sem_rd, 10);
    lprintf("ret: %d/%d PTS: %lld -> %lld\n", ret, this->ds_mpi.vd->ret, buf->pts, this->ds_mpi.vd->pts);
    //printf("PTS (%d): %f(%d) -> %d\n", this->ds_mpi.vd->ret, sh->buffered_pts[0], pts-1, this->ds_mpi.vd->pts);
    if(ret == 0 && this->ds_mpi.vd->ret && ! (this->ds_mpi.vd->ret & (1<<31))) {
      int flags = VO_BOTH_FIELDS;
      if(this->ds_mpi.vd->ret & 0x10)
	flags |= VO_INTERLACED_FLAG;
      else
	flags &= ~VO_INTERLACED_FLAG;

      img = this->stream->video_out->get_frame (this->stream->video_out,
                                        this->width, this->height,
                                        this->ratio, XINE_IMGFMT_YUY2,
                                        flags);
      if(this->ds_mpi.vd->pts)
        img->pts = this->ds_mpi.vd->pts;
      else
        img->pts = this->pts;
      this->pts = 0;

      img->bad_frame                    = 0;
      img->duration                     = this->video_step;
      if(this->ds_mpi.vd->ret & 0x02) {
        unsigned char page = this->ds_mpi.vd->ret >> 8;
        xine_fast_memcpy(img->base[0], this->ds_mpi.pagestart + page * this->ds_mpi.pagesize, this->ds_mpi.picsize);
      } else {
        xine_fast_memcpy(img->base[0], this->ds_mpi.picture, this->ds_mpi.picsize);
      }
      img->draw(img, this->stream);
      img->free(img);
    }
  }
}

/* flush gets called by video-decoder to force outputting frames (?) */
static void dshowserver_flush (video_decoder_t *this_gen) {
}

/* reset seems to be called after a seek */
static void dshowserver_reset (video_decoder_t *this_gen) {
  dshowserver_decoder_t *this = (dshowserver_decoder_t *) this_gen;
  xprintf (this->stream->xine, XINE_VERBOSITY_DEBUG, "dshowserver: called reset\n");
  if(this->decoder_initialized) {
    this->ds_mpi.vd->cmd = VD_SEEK; //'3' is cmd for seek
    sem_post(this->ds_mpi.sem_wr);
    sem_twait(this->ds_mpi.sem_rd, 10);
  }
  xprintf (this->stream->xine, XINE_VERBOSITY_DEBUG, "dshowserver: reset done\n");
}

static void dshowserver_discontinuity (video_decoder_t *this_gen) {
}

static void dshowserver_dispose (video_decoder_t *this_gen) {
  dshowserver_decoder_t *this = (dshowserver_decoder_t *) this_gen;
  if( this->decoder_initialized )  {  
    this->decoder_ok = 0;
    this->stream->video_out->close(this->stream->video_out, this->stream);
    this->ds_mpi.vd->cmd = VD_END; //'1' is cmd for terminating
    sem_post(this->ds_mpi.sem_wr);
    close(this->ds_mpi.fd);
    sem_close(this->ds_mpi.sem_wr);
    sem_close(this->ds_mpi.sem_rd);
  }
  if(this->parser)
    free_parser(this->parser);
  if(this->bih)
    free(this->bih);
  free (this);
}

static video_decoder_t *open_plugin (video_decoder_class_t *class_gen, xine_stream_t *stream) {

  dshowserver_decoder_t  *this             = (dshowserver_decoder_t *) calloc (1, sizeof (dshowserver_decoder_t));

  this->video_decoder.decode_data       = dshowserver_decode_data;
  this->video_decoder.flush             = dshowserver_flush;
  this->video_decoder.reset             = dshowserver_reset;
  this->video_decoder.discontinuity     = dshowserver_discontinuity;
  this->video_decoder.dispose           = dshowserver_dispose;

  this->stream                          = stream;
  this->class                           = (dshowserver_class_t *) class_gen;

  this->decoder_ok                      = 0;
  return &this->video_decoder;
}

#if VIDEO_DECODER_IFACE_VERSION < 19
static char *get_identifier (video_decoder_class_t *this) {
  return "dshowserver";
}

static char *get_description (video_decoder_class_t *this) {
  return "Dshowserver video decoder plugin";
}
#endif

static void dispose_class (video_decoder_class_t *this) {
  free (this);
}

static void *init_plugin (xine_t *xine, void *data) {

  dshowserver_class_t *this                = (dshowserver_class_t *) calloc (1, sizeof (dshowserver_class_t));

  this->decoder_class.open_plugin       = open_plugin;
#if VIDEO_DECODER_IFACE_VERSION < 19
  this->decoder_class.get_identifier    = get_identifier;
  this->decoder_class.get_description   = get_description;
#else
  this->decoder_class.identifier = "dshowserver";
  this->decoder_class.description = "Dshowserver video decoder plugin";
#endif
  this->decoder_class.dispose           = dispose_class;

  return this;
}

static uint32_t video_types[] = {
  BUF_VIDEO_H264,
  0
};

static const decoder_info_t dec_info_video = {
  video_types,         /* supported types */
  100                  /* priority        */
};

const plugin_info_t xine_plugin_info[] EXPORTED = {
  /* type, API, "name", version, special_info, init_function */
#if VIDEO_DECODER_IFACE_VERSION < 19
  { PLUGIN_VIDEO_DECODER, 18, "dshowserver", XINE_VERSION_CODE, &dec_info_video, init_plugin },
#else
  { PLUGIN_VIDEO_DECODER, 19, "dshowserver", XINE_VERSION_CODE, &dec_info_video, init_plugin },
#endif
  { PLUGIN_NONE, 0, "", 0, NULL, NULL }
};
